''*************************************
''*  Altair Simulator I/O System v1.2 *
''*  (C) 2007 Michael Rychlik         *
''*************************************

' v1.0 - 01 Oct 2007 - original version
' v1.1 - 09 Dec 2007 - No change just bump verion inline with other modules
' v1.2 - 09 Mar 2008 - Replaced LCD console output with PC_Text for use with PropTerminal
'                      Replaced Simple serial console input with PC_Keyboard for use with PropTerminal
' 
CON


'The following taken from the SIMH Altair simulator cfgccp.lib
'WARNING !!! Change this is altair_mem also if you change it here !!!

msize =    20          ' adjust per installed system memory (K bytes)
bioslen = $d00         ' adjust as bios changes are made (if necessary)
ccplen = $800          ' cp/m constant
bdoslen = $e00         ' cp/m constant

' cp/m image size (rounded up to next 1k boundary)
cpmlen = (ccplen + bdoslen + bioslen + $3ff) / $400
ccpph    = (msize - cpmlen) * 1024         ' ccp start address
bdosph = ccpph + ccplen                    ' bdos start address
biosph = ccpph + ccplen + bdoslen          ' bios start address

bdos     = bdosph + 6                      ' Entry point for BDOS




'The following taken from SIMH Altair simulator CBIOSX.MAC
ndisks  = 8                     ' total number of Altair disks
tracks  = 254                   ' number of tracks for regular drives
track1  = tracks+1              ' indicator for unknown track position
asecsiz = 137                   ' sector size Altair
csecsiz = $0080                 ' sector size CP/M



'       Address         Mode    Function
'       -------         ----    --------
'       selout          Out     Selects and enables controller and drive
'       statin          In      Indicates status of drive and controller
'       dskcon          Out     Controls disk function
'       secpos          In      Indicates current sector position of disk
'       dskwrit         Out     Write data
'       dskread         In      Read data

selout = 8                      ' port to select and enable controller and drive (OUT)
'       +---+---+---+---+---+---+---+---+
'       | C | X | X | X |   Device      |
'       +---+---+---+---+---+---+---+---+
'
'       C       = If this bit is 1, the disk controller selected by 'device' is
'                 cleared. If the bit is zero, 'device' is selected as the
'                 device being controlled by subsequent I/O operations.
'       X       = not used
'       Device  = value zero thru 15, selects drive to be controlled.

statin = 8                      ' port indicating status of drive and controller (IN)
'       +---+---+---+---+---+---+---+---+
'       | R | Z | I | X | X | H | M | W |
'       +---+---+---+---+---+---+---+---+
'
'       W - When 0, write circuit ready to write another byte.
'       M - When 0, head movement is allowed
'       H - When 0, indicates head is loaded for read/write
'       X - not used (will be 0)
'       I - When 0, indicates interrupts enabled (not used this simulator)
'       Z - When 0, indicates head is on track 0
'       R - When 0, indicates that read circuit has new byte to read

dskcon = 9                      ' port to control disc function (OUT)
'       +---+---+---+---+---+---+---+---+
'       | W | C | D | E | U | H | O | I |
'       +---+---+---+---+---+---+---+---+
'
'       I - When 1, steps head IN one track
'       O - When 1, steps head OUT one track
'       H - When 1, loads head to drive surface
'       U - When 1, unloads head
'       E - Enables interrupts (ignored by this simulator)
'       D - Disables interrupts (ignored by this simulator)
'       C - When 1 lowers head current (ignored by this simulator)
'       W - When 1, starts Write Enable sequence:
'           W bit on device 'statin' (see above) will go 1 and data will be read from
'           port 'dskread' until 137 bytes have been read by the controller from
'           that port. The W bit will go off then, and the sector data will be written
'           to disk. Before you do this, you must have stepped the track to the desired
'           number, and waited until the right sector number is presented on
'           device 'secpos', then set this bit.

secpos = 9                      ' port to indicate current sector position of disk (IN)
'       As the sectors pass by the read head, they are counted and the
'       number of the current one is available in this register.
'
'       +---+---+---+---+---+---+---+---+
'       | X | X |  Sector Number    | T |
'       +---+---+---+---+---+---+---+---+
'
'       X               = Not used
'       Sector number   = binary of the sector number currently under the head, 0-31.
'       T               = Sector True, is a 1 when the sector is positioned to read or write.

dskwrit = 10                    ' port to write data (OUT)
dskread = 10                    ' port to read data (IN)

'       All I/O is via programmed I/O. Each device has a status port
'       and a data port. A write to the status port can select
'       some options for the device although the simulator only
'       recognizes the reset command (0x03).
'       A read of the status port gets the port status:
'
'       +---+---+---+---+---+---+---+---+
'       | X | X | X | X | X | X | O | I |
'       +---+---+---+---+---+---+---+---+
'
'       I - A 1 in this bit position means a character has been received
'               on the data port and is ready to be read.
'       O - A 1 in this bit means the port is ready to receive a character
'               on the data port and transmit it out over the serial line.
'
'       A read to the data port gets the buffered character, a write
'       to the data port writes the character to the device.
constat = 16                    ' sio port 1 status port
condata = 17                    ' sio port 1 data port
punstat = 18                    ' sio port 2 status port
pundata = 19                    ' sio port 2 data port

' masks for disk controller (statin)
mhm     = $02                   ' head movement mask
mtzero  = $40                   ' head on track zero mask
mall    = $ff                   ' everything ok mask

' commands for disk controller (dskcon)
cstepin = $01                   ' step in command
cstepot = $02                   ' step out command
cload   = $04                   ' load head to drive surface command
cuload  = $08                   ' unload head from drive surface command
cwrseq  = $80                   ' 'start write enable sequence' command

' masks for SIO controller (constat, punstat)
mout    = $02                   ' output allowed mask

' commands for SIO controller (constat, punstat)
creset  = 3                     ' reset command

' constants for hard disk port
hdskReset      = 1                      ' command to reset controller
hdskRead       = 2                      ' read command
hdskWrite      = 3                      ' write command
hdskport       = $fd                    ' control port for simulated hard disk
firstSector    = 17                     ' first sector to load
firstTrack     = 0                      ' from this track
firstDiskAddr  = 256*firstTrack+firstSector
sectors        = (ccplen+bdoslen)/csecsiz


                                                        
VAR     
  byte ctl_if[8]      'I/O control interface 
  long Stack[100]       'Stack space for new cog
  byte Cog            'Hold ID of cog in use, if any
  byte data           'Data byte in or out
  byte port           'Port number to read/write from
  byte action

  byte altair_drive        'The Altair disk drive currently selected, -1 = none
  byte altair_track[16]    'The current track position on each Altair disk drive
  byte altair_sector       'The current sector on an altair disk
  byte altair_bidx         'Byte index into Altair disk sectors
  byte altair_write_enable 'Bytes can be written to the Altair disk when = 1
  
  byte disk_buff[512]
  long sd_block

  long d 'Debug variables
  long t
  long s
  long b
  long p
  
OBJ

  disk    : "sdspiqasm"
  kb      : "PC_Keyboard"    'You can also use Keyboard object for PS2-Keyboard                          
  TV      : "PC_Text"        'You can also use TV_Text or VGA_Text objects


   
PUB start : Success | k, x             

  {{Start new io watch process in a new cog.; return TRUE if successful}}
  Stop
  Success := (Cog := cognew(ios, @Stack) + 1)

  'Start the terminal
  TV.start(12)
  TV.str(string("Console out started",13))

  'Start the keyboard
  x := \kb.start(26, 27)
  TV.str (string("Starting keyboard... "))
  if x == 0
    TV.str (string("FAIL",13))
  else
    TV.str (string("OK",13))

  'Start the disk
  TV.str (string("Starting disk... "))
  x := \disk.start(0)
  if x == 0
    TV.str (string("FAIL",13))
  else
    TV.str (string("OK",13))

  

PUB Stop
  {{Stop watch process, if any.}}
  if Cog
    cogstop(Cog~ - 1)


PUB Active: YesNo
  {{Return TRUE if process is active, FALSE otherwise.}}
  YesNo := Cog > 0


PUB get_ctl_addr : ctl_addr     
  ctl_addr := @ctl_if





PUB ios | x
  repeat

    action := ctl_if[0]
    port   := ctl_if[1]                                 
    data   := ctl_if[2]
  
    if action == %00000011                       'An input operation ?
      case port
        constat:                                 'Console status port.
          if kb.gotkey == 0
            data := %00000010                    'Return no char ready for input and output rdy.
          else
            data := %00000011                    'Return char ready for input and output rdy.

        condata:                                 'Console data port ?
          data := kb.key                         'Return a keyboard character

        statin:                                  'Altair disk status port
          data := 0
          'if altair_write_enable == 1           'Ready to accept data ?
            'data |= %00000001
          'data |= %00000010                     'Head movement is always allowed !
          'data |= %00000100                     'Head is loaded for read/write
          if altair_track[altair_drive] <> 0     'Now on track zero ?
            data |= %01000000
          'data |= %10000000                      'New byte ready to read
        
        secpos:                                  'Altair disk sector position port
          altair_sector := altair_sector + 1
          if altair_sector > 31
            altair_sector := 0
          data := (altair_sector << 1) '| 1       'Return sector and positioned bit
          altair_bidx := 0

          'data := 0

          
        dskread:                                 'Altair disk read port 
          if altair_bidx == 0
            'Read a new block from SD card
            sd_block := (altair_drive * tracks * 32) + (altair_track[altair_drive] * 32) + altair_sector
            x := \disk.readblock(sd_block, @disk_buff)      
          data := disk_buff[altair_bidx]
          if altair_bidx < asecsiz
            altair_bidx++
      ctl_if[2] := data                          'Return the data
      ctl_if[0] := 0                             'Set flag ready for next io          

    elseif action == %00000010                   'An output operation ?
      case port
        0:                                       'Port 0 is our leds
          outa[16..23] := data                   'Ouput the port data
        condata:                                 'Console data port
          TV.out(data)
        selout:                                   'Altair disk controller select port
          if data & %10000000
            altair_drive := -1
          else
            altair_drive := data & %00001111
        dskcon:                                   'Altair disk function port
          if (data & cstepin) > 0                 'Step in command ?
            if altair_track[altair_drive] < tracks
              altair_track[altair_drive]++

          if (data & cstepot) > 0                 'Step out command ?
            if altair_track[altair_drive] > 0
              altair_track[altair_drive]--
          
          if (data & cload) > 0                   'Load head to drive surface command ?

          if (data & cuload) > 0                  'Unload head from drive surface command ?

          if (data & cwrseq) > 0                  ' 'start write enable sequence' command
            altair_bidx := 0
            altair_write_enable := 1
          
        dskwrit:                                  'Altair disk write port
          if altair_bidx < asecsiz
            disk_buff[altair_bidx++] := data
            if altair_bidx == asecsiz
              'Write the buffer to the SD card
              sd_block := (altair_drive * tracks * 32) + (altair_track[altair_drive] * 32) + altair_sector
              x := \disk.writeblock(sd_block, @disk_buff)
              altair_write_enable := 0
      ctl_if[0] := 0                               'Set flag ready for next io        
    